/*$6 ***************************************************************************
Copyright (C) <2007> <Sven Trenkel, Rene Weller, Gabriel Zachmann>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
 ******************************************************************************* */

/**
 *  \ ColUtils \
 */

/*!
 *  @file @brief Utility-functions for collision benchmarking 
 *
 *  @author Gabriel Zachmann, Sven Trenkel, Rene Weller, Gabriel Zachmann
*/

/*
 * Includes ;
 */
#include "Utils.h"
#include "Bench.h"
#include <stdlib.h>
#include <time.h>
#include <math.h>

#ifdef __sgi
#include <sys/sysmp.h>
#include <sys/pda.h>
#endif
#if defined( __sgi ) || defined( __linux )
#include <sys/times.h>
#include <sys/time.h>
#endif
#ifndef _WIN32
#include <unistd.h>
#include <alloca.h>
#include <sys/resource.h>
#endif
#include <OpenSG/OSGBaseTypes.h>
#include <OpenSG/OSGBaseFunctions.h>
#include <OpenSG/OSGGeometry.h>
#include <OpenSG/OSGGeoProperty.h>
#include <OpenSG/OSGGeoFunctions.h>
#include <OpenSG/OSGVector.h>
#include <OpenSG/OSGQuaternion.h>
#include <OpenSG/OSGMatrix.h>
#include <OpenSG/OSGTransform.h>
#include <OpenSG/OSGCamera.h>
#include <OpenSG/OSGPerspectiveCamera.h>
#include <OpenSG/OSGDrawAction.h>
#include <OpenSG/OSGSolidBackground.h>
#include <OpenSG/OSGGroup.h>
#include <OpenSG/OSGDirectionalLight.h>
#include <OpenSG/OSGViewport.h>
#include <OpenSG/OSGBoxVolume.h>
#include <OpenSG/OSGSimpleGeometry.h>
#include <OpenSG/OSGSimpleAttachments.h>
#include <OpenSG/OSGSimpleSceneManager.h>
#include <OpenSG/OSGAttachment.h>
#include <OpenSG/OSGGLUTWindow.h>
#include <OpenSG/OSGTrackball.h>
#include <OpenSG/OSGFaceIterator.h>
#include <OpenSG/OSGSimpleMaterial.h>
#include <OpenSG/OSGSceneFileHandler.h>
#include <OpenSG/OSGTriangleIterator.h>
#include <OpenSG/OSGMaterialGroup.h>
#include <OpenSG/OSGFileGrabForeground.h>

using osg:: beginEditCP;
using osg:: endEditCP;

/** ----------------------------------------------------------------------------
 *  Defines ;
 ------------------------------------------------------------------------------- */
#define max3( a, b, c ) \
        ( (a) > (b) ? (a) > (c) ? (a) : (c) : (b) > (c) ? (b) : (c) )

/* @} */


/*
 * Utility Functions ;
 */

/** ----------------------------------------------------------------------------
 *  Return the pointer to the geometry core of the node 
 ------------------------------------------------------------------------------- */
osg::GeometryPtr getGeom( const osg::NodePtr node )
{
    osg::GeometryPtr    geo = osg::GeometryPtr::dcast( node->getCore() );
    if( geo.getCPtr() == NULL )
        fprintf( stderr,
                 "getGeom: dcast failed!\nnode doesn't seem to have a geometry" );
    return geo;
}

/** ----------------------------------------------------------------------------
 *  Return AABB-Box of the geometry Node @
 ------------------------------------------------------------------------------- */
void getAABB( const osg::GeometryPtr &geom, Pnt3f &Min, Pnt3f &Max )
{
    osg::TriangleIterator   ti = geom->beginTriangles(),
    end = geom->endTriangles();
    Min = Pnt3f( ti.getPosition( 0 )[0], ti.getPosition( 0 )[1],
                 ti.getPosition( 0 )[2] );
    Max = Min;

    for( ; ti != end; ++ti )
    {
        for( int i = 0; i < 3; ++i )
        {
            Pnt3f   pos = ti.getPosition( i );
            for( int j = 0; j < 3; ++j )
            {
                if( pos[j] < Min[j] )
                    Min[j] = pos[j];
                if( pos[j] > Max[j] )
                    Max[j] = pos[j];
            }
        }
    }
}

/* -----------------------------------------------------------------------------
 *  \brief return nr of triangles of geometry node
 ------------------------------------------------------------------------------- */
unsigned countTriangles( osg::GeometryPtr geo, double *avArea )
{
    unsigned                ret = 0;
    double                  area = 0;
    Tri                     triangles;
    osg::GeoPositionsPtr    coordPtr = geo->getPositions();
    osg::TriangleIterator   ti = geo->beginTriangles(),
    ti_end = geo->endTriangles();

    for( ; ti != ti_end; ++ti )
    {
        ++ret;
        for( int j = 0; j < 3; ++j )
        {
            triangles.idx[j] = ti.getPositionIndex( j );
        }

        triangles.coordPtr = coordPtr;
        area += triangles.triArea();
    }

    if( avArea )
        *avArea = area / ret;
    return ret;
}

/* -----------------------------------------------------------------------------
 *  \brief Get triangles from a geometry pointer
 ------------------------------------------------------------------------------- */
void getTriangles( osg::GeometryPtr        geo,
                   Tri                     *triangles,
                   osg::GeoPositionsPtr    coordPtr )
{
    osg::TriangleIterator   ti = geo->beginTriangles(),
    ti_end = geo->endTriangles();
    for( unsigned i = 0; ti != ti_end; ++ti, ++i )
    {
        for( int j = 0; j < 3; ++j )
        {
            triangles[i].idx[j] = ti.getPositionIndex( j );
        }

        triangles[i].pos = i;
        triangles[i].coordPtr = coordPtr;
        triangles[i].triArea();
    }
}

/* -----------------------------------------------------------------------------
 *  \brief compares two triangles
 *  \param a,b triangles
 *  \return wether a or b has smaller Area 
 ------------------------------------------------------------------------------- */
int sortBySize( const void *a, const void *b )
{
    const Tri   *A = ( const Tri * ) a, *B = ( const Tri * ) b;
    return A->area > B->area ? -1 : A->area < B->area;
}

#define SUBDIV  0
#define GENAU

struct sBF
{
    int                     offset;
    osg::GeoPositions3fPtr  points;
    osg::GeometryPtr        lastGeo, copy;
    osg::NodePtr            root;
};

/* -----------------------------------------------------------------------------
 *  \brief Add faces to a given node. Used in the merge function
 ------------------------------------------------------------------------------- */
static void addFace( const osg::NodePtr      &node,
                     const osg::GeometryPtr  &geo,
                     const osg::FaceIterator &face,
                     void                    *data )
{
    int         len;
    struct sBF  *bf = ( struct sBF * ) data;

    if( bf->lastGeo != geo )
    {
        osg::Matrix Transform;
        getTransfomUpto( node, bf->root, Transform );
        osg::MFPnt3f * p = getPoints( geo );
        bf->lastGeo = geo;
        bf->offset = bf->points->size();
        for( int i = 0; i < p->size(); ++i )
        {
            bf->points->push_back( (*p)[i] );
            Transform.multMatrixPnt( (*bf->points->getFieldPtr())[bf->offset + i] );
        }
    }

    len = face.getLength();
    bf->copy->getLengths()->addValue( len );
    if( len == 3 )
        bf->copy->getTypes()->addValue( GL_TRIANGLES );
    else
        bf->copy->getTypes()->addValue( GL_QUADS );
    for( int i = 0; i < len; ++i )
    {
        bf->copy->getIndices()->addValue( bf->offset + face.getPositionIndex( i ) );
    }
}

/*! ----------------------------------------------------------------------------
 *  \brief  merge multiple geometry cores. This function creates a new osg node
 *          with a geometry core and fills it with all polygons of the original
 *          object that might have been stored in multiple nodes. 
 *  \param  original The original object that might consist of multiple
 *          geometry nodes and transformations.
 *  \param copy The new node that
 *          will contain the whole object in a single node.
 ------------------------------------------------------------------------------- */
void merge( const osg::NodePtr &original, osg::NodePtr &copy )
{
    struct sBF              bf;

    osg::GeometryPtr        geo;
    osg::GeoPositions3fPtr  pos;

    pos = osg::GeoPositions3f::create();
    geo = osg::Geometry::create();
    beginEditCP( geo );
    geo->setPositions( pos );
    geo->setIndices( osg::GeoIndicesUI32::create() );
    geo->setLengths( osg::GeoPLengthsUI32::create() );
    geo->setTypes( osg::GeoPTypesUI8::create() );
    geo->setTexCoords( osg::NullFC );
    geo->setNormals( osg::NullFC );
    geo->setMaterial( getGeom( findGeomNode( original ) )->getMaterial() );

    bf.copy = geo;
    bf.lastGeo = osg::NullFC;
    bf.root = original;
    bf.points = pos;

    iterFaces( original, addFace, ( void * ) &bf );

    osg::createSharedIndex( geo );
    osg::calcVertexNormals( geo );
    endEditCP( geo );

    copy = osg::Node::create();
    beginEditCP( copy );
    copy->setCore( geo );
    endEditCP( copy );
}

/* -----------------------------------------------------------------------------
 *  \brief computes a distance object for a given OpenSG object
 *  \param geo geometry of original objects
 *  \dist distance for the offset surface
 *  \return returns the offset object
 ------------------------------------------------------------------------------- */
osg::NodePtr createDistanceObject( unsigned int mode, osg::NodePtr node, double dist )
{
    
    if ( dist <= NearZero ) return node;

    osg::GeometryPtr geo = getGeom( node );
    osg::GeoPositionsPtr coordPtr = geo->getPositions()->clone();
    osg::GeoPTypesPtr typePtr = osg::GeoPTypesUI8::create();
    osg::GeoPLengthsPtr lenPtr = osg::GeoPLengthsUI32::create();
    osg::GeoIndicesPtr indexPtr = osg::GeoIndicesUI32::create();
    unsigned Pnts = coordPtr->getSize(), origPnts = coordPtr->getSize();

    if ( mode > 1 )
    {
        osg::GeometryPtr Sphere = osg::makeSphereGeo( 0, dist );
        osg::GeoPositionsPtr SphereCoord = Sphere->getPositions();
        for ( int i = 0; i < origPnts; ++i )
        {
            Pnt3f Pos = coordPtr->getValue( i );
            for ( int j = 0; j < SphereCoord->getSize(); ++j )
            {
                Vec3f SpherePos = Vec3f( SphereCoord->getValue( j ) );
                coordPtr->push_back( Pos + SpherePos );
            }
            osg::TriangleIterator ti = Sphere->beginTriangles(), ti_end = Sphere->endTriangles();
            for ( ; ti != ti_end; ++ti )
            {
                typePtr->push_back( GL_TRIANGLES );
                lenPtr->push_back( 3 );
                indexPtr->addValue( Pnts + ti.getPositionIndex( 0 ) );
                indexPtr->addValue( Pnts + ti.getPositionIndex( 1 ) );
                indexPtr->addValue( Pnts + ti.getPositionIndex( 2 ) );
            }
            Pnts += SphereCoord->getSize();
        }
    }

    osg::TriangleIterator ti = geo->beginTriangles(), ti_end = geo->endTriangles();
    for ( ; ti != ti_end; ++ti )
    {
        Pnt3f P, Q, R;
        Vec3f PQ, PR, PQPR;

        P = ti.getPosition( 0 );
        Q = ti.getPosition( 1 );
        R = ti.getPosition( 2 );
        PQ = Q - P;
        PR = R - P;
        PQPR = PQ % PR;
        PQPR.normalize();
        PQPR *= dist;

        typePtr->push_back( GL_TRIANGLES );
        lenPtr->push_back( 6 );

        coordPtr->push_back( P + PQPR );
        coordPtr->push_back( Q + PQPR );
        coordPtr->push_back( R + PQPR );

        indexPtr->addValue( Pnts );
        indexPtr->addValue( Pnts + 1 );
        indexPtr->addValue( Pnts + 2 );
        Pnts += 3;

        coordPtr->push_back( P - PQPR );
        coordPtr->push_back( Q - PQPR );
        coordPtr->push_back( R - PQPR );
        indexPtr->addValue( Pnts );
        indexPtr->addValue( Pnts + 1 );
        indexPtr->addValue( Pnts + 2 );
        Pnts += 3;

        Vec3f PQpPR = (P - Q) + (P - R);
        Vec3f QPpQR = (Q - P) + (Q - R);
        Vec3f RPpRQ = (R - P) + (R - Q);

        PQpPR.normalize();
        QPpQR.normalize();
        RPpRQ.normalize();
        coordPtr->push_back( P + PQpPR * dist );
        coordPtr->push_back( Q + QPpQR * dist );
        coordPtr->push_back( R + RPpRQ * dist );
        Pnts += 3;

        typePtr->push_back( GL_TRIANGLES );
        lenPtr->push_back( 36 );

        indexPtr->addValue( Pnts - 9 );
        indexPtr->addValue( Pnts - 8 );
        indexPtr->addValue( Pnts - 3 );
        indexPtr->addValue( Pnts - 8 );
        indexPtr->addValue( Pnts - 3 );
        indexPtr->addValue( Pnts - 2 );
        indexPtr->addValue( Pnts - 6 );
        indexPtr->addValue( Pnts - 5 );
        indexPtr->addValue( Pnts - 3 );
        indexPtr->addValue( Pnts - 5 );
        indexPtr->addValue( Pnts - 3 );
        indexPtr->addValue( Pnts - 2 );

        indexPtr->addValue( Pnts - 9 );
        indexPtr->addValue( Pnts - 7 );
        indexPtr->addValue( Pnts - 3 );
        indexPtr->addValue( Pnts - 7 );
        indexPtr->addValue( Pnts - 3 );
        indexPtr->addValue( Pnts - 1 );
        indexPtr->addValue( Pnts - 6 );
        indexPtr->addValue( Pnts - 4 );
        indexPtr->addValue( Pnts - 3 );
        indexPtr->addValue( Pnts - 4 );
        indexPtr->addValue( Pnts - 3 );
        indexPtr->addValue( Pnts - 1 );

        indexPtr->addValue( Pnts - 8 );
        indexPtr->addValue( Pnts - 7 );
        indexPtr->addValue( Pnts - 2 );
        indexPtr->addValue( Pnts - 7 );
        indexPtr->addValue( Pnts - 2 );
        indexPtr->addValue( Pnts - 1 );
        indexPtr->addValue( Pnts - 5 );
        indexPtr->addValue( Pnts - 4 );
        indexPtr->addValue( Pnts - 2 );
        indexPtr->addValue( Pnts - 4 );
        indexPtr->addValue( Pnts - 2 );
        indexPtr->addValue( Pnts - 1 );
	}

	osg::GeometryPtr newGeoPtr = osg::Geometry::create();

	beginEditCP( newGeoPtr );
	newGeoPtr->setPositions( coordPtr );
	newGeoPtr->setIndices( indexPtr );
	newGeoPtr->setLengths( lenPtr );
	newGeoPtr->setTypes( typePtr );
	newGeoPtr->setTexCoords( osg::NullFC );
	newGeoPtr->setNormals( osg::NullFC );
	newGeoPtr->setMaterial( geo->getMaterial() );
	osg::createSharedIndex( newGeoPtr );
	osg::calcVertexNormals( newGeoPtr );
	endEditCP( newGeoPtr );

	osg::NodePtr rootPtr = osg::Node::create();
	beginEditCP( rootPtr );
	rootPtr->setCore( newGeoPtr );
	endEditCP( rootPtr );

	return rootPtr;
}


/* -----------------------------------------------------------------------------
 *  \brief Computes the area of a triangle
 ------------------------------------------------------------------------------- */
double Tri::triArea()
{
    osg::Pnt3f  P, Q, R;
    osg::Vec3f  PQ, PR, PQPR;

    P = coordPtr->getValue( idx[0] );
    Q = coordPtr->getValue( idx[1] );
    R = coordPtr->getValue( idx[2] );
    PQ = Q - P;
    PR = R - P;
    PQPR = PQ % PR;
    area = PQPR.length() / 2;
    return area;
}

/*! ----------------------------------------------------------------------------
 *  \brief Combine all transformation matrices between two nodes in the graph
 *  @param cur lowest node of the graph 
 *  @param upto highest node in the graph
 *  @param result the resulting transformation matrix
 ------------------------------------------------------------------------------- */
void getTransfomUpto( const osg::NodePtr  &cur,
                      const osg::NodePtr  &upto,
                      osg::Matrix         &result )
{
    osg::TransformPtr   transform = osg::TransformPtr::dcast( cur->getCore() );
    const osg::NodePtr  Parent = cur->getParent();
    if( cur == upto || cur == osg::NullFC )
    {
        result.setIdentity();
        return;
    }

    if( Parent != osg::NullFC )
        getTransfomUpto( Parent, upto, result );
    if( transform == osg::NullFC )
        return;
    result.mult( transform->getMatrix() );
}

/*! ----------------------------------------------------------------------------
 *  \brief Return the pointer to the multi-field of the points @warning Don't use the
 *  MFPnt3f pointer to modify the geometry! OpenSG will not notice the changes!
  ------------------------------------------------------------------------------- */

osg::MFPnt3f *getPoints ( const osg::GeometryPtr geo )
{
    const osg::GeoPositions3fPtr    pos = osg::GeoPositions3fPtr::dcast
        ( geo->getPositions() );
    if( pos.getCPtr() == NULL )
        return NULL;
    return pos->getFieldPtr();
}

/*! ----------------------------------------------------------------------------
 *  \brief Find the first node that has a geometry
 *  \param node root of subtree to be searched
 *  \return The node having a geometry, or osg::NullFC. Make a
 *      depth-first traversal of the subtree starting at \a node, and return the
 *      first node that has a geometry.
 ------------------------------------------------------------------------------- */
osg::NodePtr findGeomNode( const osg::NodePtr node )
{
    osg::GeometryPtr    geo = osg::GeometryPtr::dcast( node->getCore() );
    if( geo != osg::NullFC )
    {
        if( geo->getMaterial() == osg::NullFC )
        {
            osg::MaterialGroupPtr   mg = osg::NullFC;

            for( osg::NodePtr n = node; n != osg::NullFC; n = n->getParent() )
            {
                mg = osg::MaterialGroupPtr::dcast( n->getCore() );
                if( mg != osg::NullFC )
                {
                    beginEditCP( geo, osg::Geometry::MaterialFieldMask );
                    geo->setMaterial( mg->getMaterial() );
                    endEditCP( geo, osg::Geometry::MaterialFieldMask );
                    break;
                }
            }
        }

        return node;
    }

    for( unsigned int i = 0; i < node->getNChildren(); i++ )
    {
        osg::NodePtr    childnode = findGeomNode( node->getChild( i ) );
        if( childnode != osg::NullFC )
            return childnode;
    }

    return osg::NullFC;
}

/*! ----------------------------------------------------------------------------
 *  \brief Calls a function for every face in the scenegraph
 *  \param node root of the graph
 *  \param callback the function to call
 *  \param data whatever you like
 ------------------------------------------------------------------------------- */
void iterFaces( const osg::NodePtr  &node,
                void (*callback)
                    (
                    const osg::NodePtr &,
                const osg::GeometryPtr &,
                const osg::FaceIterator &,
                void *
            ),
                void *data )
{
    osg::GeometryPtr    geo = osg::GeometryPtr::dcast( node->getCore() );
    for( unsigned int i = 0; i < node->getNChildren(); i++ )
    {
        iterFaces( node->getChild( i ), callback, data );
    }

    if( geo == osg::NullFC )
        return;

    for( osg::FaceIterator fi = geo->beginFaces(); fi != geo->endFaces(); ++fi )
    {
        callback( node, geo, fi, data );
    }
}

/*! ----------------------------------------------------------------------------
 *  Affine combination of two points 
 *  \param pnt1,pnt2 points 
 *  \return @a pnt1*c1+ @a pnt2*c2. 
 *  \pre \a c1 + \a c2 = 1!
 ------------------------------------------------------------------------------- */
osg::Pnt3f lincomb( float               c1,
                    const osg::Pnt3f    &pnt1,
                    float               c2,
                    const osg::Pnt3f    &pnt2 )
{
    osg::Pnt3f  result;
    for( unsigned int i = 0; i < 3; i++ )
        result[i] = c1 * pnt1[i] + c2 * pnt2[i];
    return result;
}

/*! ----------------------------------------------------------------------------
 *  Load a file, find the node bearing geometry, and scale it. 
 *  \param filename name of file 
 *  \param node the geometry node (out) The geometry is
 *      scaled/translated such that it fits inside the [-1,+1] box.
 ------------------------------------------------------------------------------- */
osg::NodePtr loadGeom( const char *filename )
{
    osg::NodePtr    filescene, node;
    osg::Pnt3f      low, high;
    
    /* load file */
    filescene = osg::SceneFileHandler::the().read( filename );
    if( filescene == osg::NullFC )
    {
        fprintf( stderr, "\nCouldn't open file %s\n", filename );
        exit( -1 );
    }

    merge( filescene, node );

    /* scale geom */
    node->updateVolume();
    node->getVolume().getBounds( low, high );
    osg::Pnt3f      c = lincomb( 0.5, low, 0.5, high );
    osg::Pnt3f      d = lincomb( 0.5, high, -0.5, low );
    /* d=0.5*(high-low), is actually a vector */
    float           s = max3( d[0], d[1], d[2] );
    osg::MFPnt3f * points = getPoints( getGeom( node ) );
    beginEditCP( node );
    for( unsigned int i = 0; i < points->size(); i++ )
        for( unsigned int j = 0; j < 3; j++ )
            ( *points )[i][j] = ( (*points)[i][j] - c[j] ) / s;
    endEditCP( node );
    node->updateVolume();
    return node;
}

/* -----------------------------------------------------------------------------
 *  \brief divides strings into tokens
 *  \param buf original string
 *  \param divider string that divides the tokens
 *  \return first token
 ------------------------------------------------------------------------------- */
char *getToken( char **buf, const char *divider )
{
    char    *ret;
    if( !buf || !*buf )
        return 0;
    while( **buf && strchr( divider, **buf ) )
        ++ *buf;
    if( !**buf )
        return *buf;
    ret = *buf;
    while( **buf && !strchr( divider, **buf ) )
        ++ *buf;
    if( **buf )
    {
        *( (*buf)++ ) = 0;
        while( **buf && strchr( divider, **buf ) )
            ++ *buf;
    }

    return ret;
}

/*! ----------------------------------------------------------------------------
 *  \brief Create geometry
 *  \param type the type of geometry to create
 *  \param complexity determines number of polygons \param node, geom output Creates a
 *      geometry and a node. The node will not be added to the root. The number of
 *      polygons is usually something like complexity^2 .
 ------------------------------------------------------------------------------- */
osg::NodePtr createGeom( GeometryTypeE type, unsigned int complexity )
{
    char                    name[1000];
    static unsigned int     ngeom = 0;
    static float            red = 1.0, green = 0.2, blue = 1.0;
    osg::NodePtr            node;
    osg::GeometryPtr        geom;

    osg::SimpleMaterialPtr  mat = osg::SimpleMaterial::create();
    osg::NamePtr            name_attachm = osg::Name::create();
    switch( type )
    {
        case OBJ_PLANES:
            {
                node = osg::makePlane( 1, 1, complexity, complexity );

                /* move one point, so that convex hull can be computed */
                osg::MFPnt3f * points = getPoints( getGeom( node ) );
                beginEditCP( node );
                ( *points )[0][2] += 0.3;
                ( *points )[points->size() - 1][2] -= 0.3;
                endEditCP( node );
                sprintf( name, "pl_%u_%u", complexity, ngeom );
                break;
            }

        case OBJ_SPHERES:
            complexity = static_cast< unsigned int >( logf( complexity * complexity ) );
            node = osg::makeSphere( complexity, 0.5 );
            sprintf( name, "sh_%u_%u", complexity, ngeom );
            break;

        case OBJ_TORUS:
            node = osg::makeTorus( 0.35, 0.5, complexity + 1, complexity + 1 );
            sprintf( name, "to_%u_%u", complexity, ngeom );
            break;

        default:
            fputs( "createGeom: BUG: type is out of range!\n", stderr );
            exit( -1 );
    }

    ngeom++;

    if( node == osg::NullFC )
    {
        fprintf( stderr, "interactive: createGeom failed!\n" );
        exit( -1 ); /* makes no sense to continue */
    }

    geom = getGeom( node );

    /* set material */
    mat->setDiffuse( osg::Color3f( red, green, blue ) );
    mat->setAmbient( osg::Color3f( 0.5, 0.5, 0.5 ) );
    mat->setSpecular( osg::Color3f( 0.5, 0.5, 0.5 ) );
    mat->setShininess( 20 );

    /* change color for next createGeom call (doesn't work too often) */
    red -= 0.2;
    green += 0.2;
    blue -= 0.2;

    geom->setMaterial( mat );

    /* set name */
    name_attachm->getFieldPtr()->getValue().assign( name );
    node->addAttachment( name_attachm );
    return node;
}

/*! ----------------------------------------------------------------------------
 *  \brief Convert a rotation given by axis & angle to a matrix. 
 *  \param a axis (unit vector) 
 *  \param d angle (radians) \return m rotation matrix 
 *  \pre - The matrix will be used via "vector * matrix". - The axis must be a unit
 *      vector. The axis/angle have the meaning: "rotate about the axis with angle
 *      deg" The right-hand rule is used. The matrix will be of the form \verbatim
 *      / 0 \ | Rot 0 | | 0 | \0 0 0 1 \endverbatim (Source: W. Schindler)
 *  \implementation Ported from Y/conversion.c 
 *  \todo d = -d; kann man wahrscheinlich wieder rauswerfen, wenn man unten die Transposition
 *      auch entfernt.
 ------------------------------------------------------------------------------- */
osg::Matrix axisToMat( const osg::Vec3f &a, float d )
{
    float       ca, eca, sa, h;
    float       m[3][3];
    osg::Matrix res;

    d = -d; /* rest of code uses left-hand rule */
    ca = cos( d );
    sa = sin( d );
    eca = 1 - ca;

    h = eca * a[0];
    m[0][0] = h * a[0] + ca;
    m[0][1] = m[1][0] = h * a[1];
    m[0][2] = m[2][0] = h * a[2];

    h = eca * a[1];
    m[1][1] = h * a[1] + ca;
    m[1][2] = m[2][1] = h * a[2];

    m[2][2] = eca * a[2] * a[2] + ca;

    h = a[2] * sa;
    m[0][1] -= h;
    m[1][0] += h;

    h = a[1] * sa;
    m[0][2] += h;
    m[2][0] -= h;

    h = a[0] * sa;
    m[1][2] -= h;
    m[2][1] += h;

    res.setValue( m[0][0], m[1][0], m[2][0], 0, m[0][1], m[1][1], m[2][1], 0,
                  m[0][2], m[1][2], m[2][2], 0, 0, 0, 0, 1 );
    return res;
}

/* -----------------------------------------------------------------------------
 *  \brief Multiply two matrices
 ------------------------------------------------------------------------------- */
osg::Matrix operator*( const osg::Matrix &m1, const osg::Matrix &m2 )
{
    osg::Matrix res( m1 );
    res.mult( m2 );
    return res;
}


/**
 *  @name Timers, timing, sleeping, etc.
 */

/* @{ */

/*! ----------------------------------------------------------------------------
 *  Get the user time in milliseconds 
 *  \return Time in milliseconds. The time is
 *  the user time of the process (and all children) since the process started.
 *  \implementation Uses \c getrusage under Unix, and \c GetProcessTimes under
 *  Windoze. Warning: Under Windoze, clock() returns wall-clock time, *not*
 *  user time! (contrary to what the MSDN documentation says!)
 ------------------------------------------------------------------------------- */
float getTime( void )
{
#ifdef _WIN32
    unsigned long long  utime;
    FILETIME            creation_time, exit_time, kernel_time, user_time;
    bool                success = GetProcessTimes( GetCurrentProcess(),
                                                   &creation_time, &exit_time,
                                                   &kernel_time, &user_time );
    if( !success )
        return 0.0;
    utime =
        (
            static_cast< unsigned long long >( user_time.dwHighDateTime ) <<
            32
        ) +static_cast< unsigned >( user_time.dwLowDateTime );
    return static_cast< float >( utime ) / 1.0e4;

#else
    struct rusage   ru;
    getrusage( RUSAGE_SELF, &ru );
    return ru.ru_utime.tv_sec * 1000.0 + ru.ru_utime.tv_usec / 1000.0;
#endif
}

/**
 *  @class NanoTimer Timer with nanoseconds resolution. The units of this timer
 *  are always nanoseconds, but on some platforms, the actual resolution might
 *  be less (microseconds or even less). Where implemented, this class uses the
 *  high-speed, high-performance, hardware timers/counters. Otherwise, we just
 *  use @a gettimeofday(), which, at least on Linux/Pentium, has microsecond
 *  resolution. Currently, the time is @e wall-clock time, not user time! @see
 *  http://www.ncsa.uiuc.edu/UserInfo/Resources/Hardware/IA32LinuxCluster/Doc/timing
 *  .html - cedar.intel.com/software/idap/media/pdf/rdtscpm1.pdf
 *  /raphael/knowledge/programming/c/code-snippets/cpu_clock_timer.c @todo Try
 *  to estimate the overhead of a function call to start() or elapsed(). Use
 *  PAPI (http://icl.cs.utk.edu/projects/papi/) or the "high resolution timers
 *  project" (http://high-res-timers.sourceforge.net/) when they become widely
 *  available (without kernel patches).
 */

/** ----------------------------------------------------------------------------
 *  Create a new timer with nanoseconds resolution Saves the current time
 *  stamp, too. @sideeffects See @a checkFrequency().
 ------------------------------------------------------------------------------- */
NanoTimer::NanoTimer( void )
{
    if( !M_FrequencyChecked )
        checkFrequency();
    start();
}

/** ----------------------------------------------------------------------------
 *  Save the current time (stamp) in the timer
 ------------------------------------------------------------------------------- */
void NanoTimer::start( void )
{
    m_time_stamp = getTimeStamp();
}

/** ----------------------------------------------------------------------------
 *  Return the time elapsed since the last @a start() in nanoseconds.
 ------------------------------------------------------------------------------- */
double NanoTimer::elapsed( void ) const
{
    unsigned long long int  now = getTimeStamp();
    if( M_Use_High_Frequ )
        return static_cast< double >( now - m_time_stamp ) / ( M_GHz * 1000000.0 );
    else
        return static_cast< double >( now - m_time_stamp ) * 1000.0;
}

/**
 *  Return current time stamp Should be used only internally by this class,
 *  because the meaning of the return value is different depending on whether
 *  @a M_Use_High_Frequ is true or not. @warning We don't check whether
 *  checkFrequency succeded! If it failed, the times will be bogus, or, in the
 *  worst case, the program might even crash.
 */
#if defined( __GNUC__ )

long long unsigned int NanoTimer::getTimeStamp( void )
{
#if defined( __pentiumpro )
    unsigned long long int  timestamp;
    __asm__ volatile( "rdtsc" : "=A"(timestamp) : );
    return timestamp;

#elif defined( __itanium )
    unsigned long long int  timestamp;
    __asm__ volatile( "mov %0=ar.itc" : "=r"(timestamp)::"memory" );
    return timestamp;
#else

    // should not happen
    unsigned long long int timestamp;
    __asm__ volatile ("rdtsc" : "=A" (timestamp) : );
    //printf ( "Get Timestamp: %llu\n",  timestamp);
    return timestamp;
#endif
}

#elif defined( _WIN32 )

long long unsigned int NanoTimer::getTimeStamp( void )
{
    LARGE_INTEGER   stamp;
    QueryPerformanceCounter( &stamp );
    return stamp.QuadPart;
}

#elif defined( __ICC )
__declspec( cpu_dispatch(pentium_iii, pentium) )

long long unsigned int NanoTimer::getTimeStamp( void )
{

    /* empty stub, which icc uses to create CPU dispatcher */
}

__declspec( cpu_specific(pentium_iii) )

long long unsigned int NanoTimer::getTimeStamp( void )
{
    unsigned long long int  timestamp;
    __asm__ volatile( "rdtsc" : "=A"(timestamp) : );
    return timestamp;
}

__declspec( cpu_specific(pentium) )

long long unsigned int NanoTimer::getTimeStamp( void )
{
    return time() * 1000;
}

#else /* all platforms not yet implemented */

long long unsigned int NanoTimer::getTimeStamp( void )
{
    return time() * 1000;
}
#endif

/** ----------------------------------------------------------------------------
 *  Determine the clock frequency of the CPU Try to determine the speed at
 *  which the CPUs time-stamp counter is running (on those plastforms where
 *  NanoTimer uses this counter), and whether or not there is a high-frequency
 *  counter at all. @pre Assumes that all CPUs in one system run with the same
 *  clock frequency! @sideeffects Class variable M_GHz. @todo - SGI
 *  @implementation Under Linux, we parse /proc/cpuinfo (if __pentiumpro was
 *  defined at compile time);
 *  under Windoze, we use QueryPerformanceFrequency. @see getTimeStamp(). The
 *  #ifdef's must match!
 ------------------------------------------------------------------------------- */
void NanoTimer::checkFrequency( void )
{
    M_GHz = 0.001;
    M_Use_High_Frequ = false;

#if defined( __pentiumpro ) || defined( __ICC ) || defined( __GNUC__ )
    FILE    *cpuinfo = fopen( "/proc/cpuinfo", "r" );
    if( !cpuinfo )
    {
        perror( "fopen(/proc/cpuinfo)" );
        fputs( "NanoTimer: Can't determine CPU frequency!\n", stderr );
    }
    else
    {
        char    cpuinfo_buf[1000];

        while( fgets( cpuinfo_buf, sizeof(cpuinfo_buf), cpuinfo ) )
        {
            if( !strncasecmp( cpuinfo_buf, "cpu mhz", 7 ) )
            {
                double  mhz;
                int     assigned;

                assigned = sscanf( cpuinfo_buf, "%*s %*s : %lf", &mhz );

                if( assigned != 1 )
                    fputs( "NanoTimer: Can't parse /proc/cpuinfo !\n", stderr );
                else
                {
                    M_GHz = mhz / 1000.0;
                    M_Use_High_Frequ = true;
                }
                break;
            }
        }

        fclose( cpuinfo );
    }

#elif defined( _WIN32 )
    LARGE_INTEGER   hz;
    bool            use_high_frequ = QueryPerformanceFrequency( &hz );
    if( !use_high_frequ )
        fputs( "NanoTimer: Can't use high-frequency counter!\n", stderr );
    else
    {
        M_GHz = hz.QuadPart / 1.0E9;
        M_Use_High_Frequ = true;
    }
#endif
    if( M_Use_High_Frequ )
        printf( "NanoTimer: found %lf GHz\n", M_GHz );
    else
    {
        puts( "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" );
        puts( "NanoTimer: will be bogus !!" );
        puts( "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" );
    }

    M_FrequencyChecked = true;
}

/** ----------------------------------------------------------------------------
 *  Tells whether or not the NanoTimer use the high frequency counter @warning
 *  Valid only after the first NanoTimer has been created!
 ------------------------------------------------------------------------------- */
bool NanoTimer::usesHighFrequ( void )
{
    return M_Use_High_Frequ;
}

/** ----------------------------------------------------------------------------
 *  Returns the frequency (resolution) of the counter in GHz @warning Valid
 *  only if @a NanoTimer::usesHighFrequ() = @a true !
 ------------------------------------------------------------------------------- */
double NanoTimer::frequ( void )
{
    return M_GHz;
}

bool NanoTimer::    M_Use_High_Frequ = false;
double NanoTimer::  M_GHz = 0.001;
bool NanoTimer::    M_FrequencyChecked = false;
