/*$6 ***************************************************************************
Copyright (C) <2007> <Sven Trenkel, Rene Weller, Gabriel Zachmann>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
 ******************************************************************************* */
/**
 *  \ Col \
 */

/**
 *  @class Wrapper for BoxTree and DopTree
 *  @author Sven Trenkel, Rene Weller, Gabriel Zachmann
 */

/*
 * Includes ;
 */
 

#include <stdio.h>

#include <OpenSG/OSGConfig.h>
#include <OpenSG/OSGBaseTypes.h>
#include <OpenSG/OSGNode.h>
#include <OpenSG/OSGVector.h>
#include <OpenSG/OSGMatrix.h>
#include <OpenSG/OSGGeometry.h>
#include <OpenSG/OSGGeoProperty.h>
#include <OpenSG/OSGGeoFunctions.h>
#include <OpenSG/OSGTriangleIterator.h>
#include <OpenSG/OSGBoxVolume.h>

#include <Collision.h>
#include <ColExceptions.h>
#include <ColObj.h>
#include <ColUtils.h>

#include "../Bench.h"

using osg::Vec3f;
using osg:: Pnt3f;

/*
 * Forward Declarations ;
 */
 
static col::ColObj **dopObjs, **boxObjs;
static col::CollisionPipeline * pipeline;

static void makeDopObjects( int num, osg::NodePtr node[] );
static void makeBoxObjects( int num, osg::NodePtr node[] );
static int  doDopTree( int num, osg::Matrix m[] );
static int  doBoxTree( int num, osg::Matrix m[] );
static void shutdownBox();
static void shutdownDop();

static tAPI dopAPI = { makeDopObjects, doDopTree, NULL, shutdownDop };

static tAPI boxAPI = { makeBoxObjects, doBoxTree, NULL, shutdownBox };


/* -----------------------------------------------------------------------------
 *  \brief Init module and register it to the benchmarking application
 ------------------------------------------------------------------------------- */
#ifdef _WIN32
extern "C"
__declspec( dllexport )
std::vector<Algorithm *> *
initModule ( void )
#else
extern "C"
{
std::vector<Algorithm *>    *initModule( void );
}
std::vector<Algorithm *>    *
initModule ( void )
#endif
{
    std::vector<Algorithm *>    *ret = new std::vector < Algorithm * > ();
    ret->clear();
    Algorithm   *boxtree = new Algorithm();
    boxtree->shortName = "bx";
    boxtree->name = "BoxTree";
    boxtree->api = boxAPI;
    ret->push_back( boxtree );

    Algorithm   *doptree = new Algorithm();
    doptree->shortName = "do";
    doptree->name = "DopTree";
    doptree->api = dopAPI;
    ret->push_back( doptree );

    pipeline = new col::CollisionPipeline( "pipe", 1 );
    return ret;
}


/* -----------------------------------------------------------------------------
 *  \brief Some special functions needed for this library
 ------------------------------------------------------------------------------- */
static col::MatrixCell * dopCell, *boxCell;

static col::ColObj *addDopObject( osg::NodePtr node )
{
    col::ColObj * obj;
    osg::GeometryPtr    geom = col::getGeom( node );

    obj = new col::ColObj( geom, node, false, false, false, col::ALGO_DOPTREE, NULL );
    return obj;
}
static col::ColObj *addBoxObject( osg::NodePtr node )
{
    col::ColObj * obj;
    osg::GeometryPtr    geom = col::getGeom( node );

    obj = new col::ColObj( geom, node, false, false, false, col::ALGO_BOXTREE, NULL );
    return obj;
}

// -----------------------------------------------------------------------------
// *  dummy class in order to implement operator () so we can create instances
// -----------------------------------------------------------------------------
class benchCallback : public col::Callback
{
public:
    benchCallback( osg::NodePtr            obj1,
                   osg::NodePtr            obj2,
                   col::LevelOfDetectionE  level );
    virtual void operator () ( const col::Data *data ) throw();
};

/* -----------------------------------------------------------------------------
 *  \brief This function is called when a collision occurs
 *         Without functionality in our benchmark
 ------------------------------------------------------------------------------- */

benchCallback::benchCallback( osg::NodePtr            obj1,
                              osg::NodePtr            obj2,
                              col::LevelOfDetectionE  level ) : col::Callback( obj1, obj2, false, false, level )
{

    //*
    // *  if (all_collisions)
    //
}

static bool                     c;

void benchCallback::operator    ( ) ( const col::Data *data )
throw()
{
    puts( "bench: BUG: coll. callback has been called! (should not happen)" );
}
static col::Callback * DopCB, *BoxCB;


/* -----------------------------------------------------------------------------
 *  \brief Make the OSG-nodes collidable
 *  \param num # of objects
 *  \param Node OSG-Nodes of the objects
 ------------------------------------------------------------------------------- */
static void makeDopObjects( int num, osg::NodePtr node[] )
{
    dopObjs = new col::ColObj *[num + 1];
    dopObjs[num] = NULL;
    for( int i = 0; i < num; ++i )
    {
        dopObjs[i] = addDopObject( node[i] );
    }

    DopCB = new benchCallback( node[0], node[1], col::LEVEL_EXACT );
    dopCell = new col::MatrixCell( dopObjs[0], dopObjs[1] );
    dopCell->addCallback( DopCB );
}

/* -----------------------------------------------------------------------------
 *  \brief Clears all memory used by the DopTrees.
 ------------------------------------------------------------------------------- */
static void shutdownDop()
{
    if ( dopCell )
    {
        delete dopCell;
        dopCell = NULL;
    }
    if ( DopCB )
    {
        delete DopCB;
        DopCB = NULL;
    }
    if ( !dopObjs ) return;
    for (int i = 0; dopObjs[i]; ++i)
    {
        delete dopObjs[i];
    }
    delete[] dopObjs;
    dopObjs = NULL;
}

/* -----------------------------------------------------------------------------
 *  \brief Make the OSG-nodes collidable
 *  \param num # of objects
 *  \param Node OSG-Nodes of the objects
 ------------------------------------------------------------------------------- */
static void makeBoxObjects( int num, osg::NodePtr node[] )
{
    boxObjs = new col::ColObj *[num + 1];
    boxObjs[num] = NULL;
    for( int i = 0; i < num; ++i )
    {
        boxObjs[i] = addBoxObject( node[i] );

        //
        // pipeline->makeCollidable(node[i]);
        //
    }

    BoxCB = new benchCallback( node[0], node[1], col::LEVEL_EXACT );
    boxCell = new col::MatrixCell( boxObjs[0], boxObjs[1] );
    boxCell->addCallback( BoxCB );

    //
    // pipeline->addCallback(BoxCB);
    //
}

/* -----------------------------------------------------------------------------
 *  \brief Clears all memory used by the BoxTrees.
 ------------------------------------------------------------------------------- */
static void shutdownBox()
{
    if ( boxCell )
    {
        delete boxCell;
        boxCell = NULL;
    }
    if ( BoxCB )
    {
        delete BoxCB;
        BoxCB = NULL;
    }
    if ( !boxObjs ) return;
    for (int i = 0; boxObjs[i]; ++i)
    {
        delete boxObjs[i];
    }
    delete[] boxObjs;
    boxObjs = NULL;
}

/* -----------------------------------------------------------------------------
 *  \brief Collision check with DopTree
 *  \param m transformation matrix of moving object
 *  \param i number of transformed object (has no effect in this version)
 ------------------------------------------------------------------------------- */
static int doDopTree( int num, osg::Matrix m[] )
{
    static unsigned cycle = 0;

    // update toworld matrix stored inside ColObj's
    cycle++;
    dopObjs[0]->hasMoved( cycle );
    dopObjs[1]->hasMoved( cycle );

    // perform coll. det., without convex hull pre-check
    return dopCell->check( false );
}

/* -----------------------------------------------------------------------------
 *  \brief Collision check with BoxTree
 *  \param m transformation matrix of moving object
 *  \param i number of transformed object (has no effect in this version)
 ------------------------------------------------------------------------------- */
static int doBoxTree( int num, osg::Matrix m[] )
{
    static unsigned cycle = 0;

    // update toworld matrix stored inside ColObj's
    cycle++;
    boxObjs[0]->hasMoved( cycle );
    boxObjs[1]->hasMoved( cycle );

    // perform coll. det., without convex hull pre-check
    c = false;
    c = /* pipeline */ boxCell->check( false );
    return c;
}
